# Autogenerated by Astropy's setup.py on 2018-04-23 18:45:18
from __future__ import unicode_literals
import datetime

version = "2.0.6"
githash = "749ee4a01a86f74f8ffe5d837941f6b17f595bbb"


major = 2
minor = 0
bugfix = 6

release = True
timestamp = datetime.datetime(2018, 4, 23, 18, 45, 18)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
